#!/usr/bin/perl
# -*- mode: CPerl; -*-

# Copyright (C_) 2011 Timothe Litt <litt at acm dot org>
# Full text at end of module.

# Installer for *wiki Tasks Framework

use warnings;
use strict;

use Cwd qw/realpath getcwd/;
use File::Basename;
use File::Spec;
use FindBin;
use Getopt::Std;
   $Getopt::Std::STANDARD_HELP_VERSION = 1;
use POSIX qw/lchown/;

our $VERSION = '3.0-001';
sub HELP_MESSAGE;

my %cliOptions;
my $sws = 'fvnb:s:t:Zzpu:U:m:';
getopts( $sws, \%cliOptions );

my $verbose = 0;
$verbose = 1 if( $cliOptions{v} );

my $rootdir;
my $bindir = realpath( $cliOptions{b} || 'bin/');
die "No bin directory specified/available\n" unless( $bindir );

my $libdir;
my $toolsdir = realpath( $cliOptions{t} || 'tools/'); # Foswiki will over-ride
my $inidir = realpath( $cliOptions{s} || (-d '/etc/init.d'? '/etc/init.d/' : '-'));
undef $inidir if $inidir =~ m!/-$!;

my $permissiveOK = $cliOptions{p};
our $fileUmask;
our $dirUmask;

my $Wikiname = -r "$bindir/setlib.cfg" && 
                  (system( "grep -q '\$foswikiLibPath' $bindir/setlib.cfg >/dev/null 2>&1" ) == 0) ? 'Foswiki' : 'TWiki';
unless( $Wikiname ) {
    print STDERR "No wiki detected in $bindir\n";
    exit 1;
}
my $wikiname = lc $Wikiname;
my $wikiver;

my $wikilib;
my $tasklib;
my $cfg;
#my $task;

my $selinux = !$cliOptions{z} && !system( "selinuxenabled" );
if( $cliOptions{Z} && !$selinux ) {
    print STDERR "SeLinux requested but not enabled (or both -Z and -z specified)\n";
    exit 1;
}
my $perms;

# Cached user and group name => id mappings

my( %users, %groups);

# Distribution (or local) mapping file

my $mapfile = $cliOptions{m};

# Selinux types (no defaults)
our %selinuxTypeMap;

# Optional SeLinux user type to apply (Default is unchanged)

our $selinuxUser;

# Known user and groups (defaults here)

our %userGroupMap = (
    rootUser => 'root',
    rootGroup => 'root',
    webserverUser => 'apache',
    webserverGroup => 'apache',
		 );

# File and directory permission maps (default to identity mapping)

our( %filePermMap, %dirPermMap );

my $cmd = shift || '';

if( $cmd eq 'install' ) {
    exit( install( @ARGV ) );
} elsif( $cmd eq 'uninstall' ) {
    exit( uninstall( @ARGV ) );
} else {
    print STDERR ( "Unknown command $cmd, try --help\n" );
    exit 1;
}

sub install {
    return 1 unless( checkwiki() && checkinit( ) );

    # Read the manifest mapping data

    $mapfile = getmappings( $mapfile, "$wikilib/Plugins/TasksPlugin/MANIFEST.map", "$libdir/MANIFEST.map" );

    # Apply umasks from command line, or use default from map file, or hardcoded

    if( exists $cliOptions{u} ) {
	$fileUmask = $cliOptions{u} =~ /^0/? oct($cliOptions{u}) : $cliOptions{u};
    }else {
	$fileUmask = 007 unless( defined $fileUmask );
    }
    if( exists $cliOptions{U} ) {
	$dirUmask = $cliOptions{U} =~ /^0/? oct($cliOptions{U}) : $cliOptions{U};
    }else {
	$dirUmask = 006 unless( defined $dirUmask );
    }

    # Read the manifest for permissions/ownership/selinux information

    if( 1 ) {
	$perms = getperms( "$wikilib/Plugins/TasksPlugin/MANIFEST" );
    } else {
	# For entire installation

	require File::Find;
	File::Find->import qw//;
	$perms = $File::Find::name;
	$perms = {};
	File::Find::find( {
			   wanted => sub {
			       return unless /^MANIFEST$/;
			       $perms = { %$perms,
					  %{ getperms( $File::Find::name ) },
					};
			   }, follow => 1, follow_skip => 2, }, $rootdir );
    }
    print "Installing framework\n";

    unless(
	   # Link for cgi-initiated start
	   slink( "$tasklib/Daemon", "$bindir/TaskDaemon" ) && 
	   ( !$inidir || initlink( $bindir, $inidir ) )
	  ) {
	print STDERR "Setup failed\n";
	exit 1;
    }

    # Apply permissions, ownership and security contexts

    for my $file (keys %$perms) {
	setperms( $rootdir, $file );
    }

    #    tick_$wikiname => .bak, and link.?
    # The idea here is to slip under tick_twiki and let its crontab entry start the daemon if stopped.
    # Not sure I want to do this.

    if( $inidir ) {
	my $initname = initname();

	if( $cliOptions{n} ) {
	    print "Would enable autostart & start $initname\n";
	} else {
	    print "Enabling $initname autostart\n";
	    unless( system( "chkconfig --add $initname" ) == 0 ) {
		print STDERR "Failed to enable $initname autostart\n";
		return 1;
	    }
	    print "Starting $initname\n";
	    unless( system(  "$inidir/$initname start" ) == 0 ) {
		print STDERR "Failed to start $initname\n";
		return 1;
	    }
	}
    }

    # Success
    return 0;
}

sub uninstall {
    return 1 unless( checkwiki( ) && checkinit( ) );

    my $initname = initname() if $inidir;

    if( $inidir ) {
	if( $cliOptions{n} ) {
	    print "Would disable autostart & stop $initname\n";
	} else {
	    print "Disabling $initname autostart\n";
	    unless( system( "chkconfig --del $initname" ) == 0 ) {
		print STDERR "Failed to disable $initname autostart\n";
	    }
	    print "Stopping $initname\n";
	    unless( system( "$inidir/$initname stop" ) == 0 ) {
		print STDERR "Failed to stop $initname\n";
	    }
	}
    }

    unless(
	   sunlink( "$bindir/TaskDaemon" ) && 
	   ( !$inidir || sunlink( realpath( "$inidir/$initname" ) . '_bin' ) ) &&
	   ( !$inidir || sunlink( realpath( "$inidir/$initname" ) . '_init' ) ) &&
	   ( !$inidir || sunlink( "$inidir/$initname" ) )
	  ) {
	print STDERR "Uninstall failed\n";
	exit 1;
    }

    return 0;
}

sub initname {

    # Name by which startup knows this wiki

    my $inilink = "$tasklib/Daemon" . '_init';
    if( -l $inilink ) {
	return (fileparse( readlink( $inilink ) ))[0];
    }

    # Default to primary (first) if no backlink

    return "${Wikiname}TaskDaemon";
}

sub initlink {
    my( $bindir, $inidir ) = @_;

    # Create startup (system boot) links for init
    # Each installed wiki can have one, so this requires a bit of work.

    return 1 unless $inidir;

    my $ininame;
    my $inilink = "$tasklib/Daemon" . '_init';
    if( -l $inilink ) {
	$ininame = (fileparse( readlink( $inilink ) ))[0];
    } else {
	my $suffix = '';
	my $seq = 0;
	while( -e "$inidir/${Wikiname}TaskDaemon$suffix" && 
	       realpath( "$inidir/${Wikiname}TaskDaemon$suffix" ) ne
	       realpath( "$tasklib/Daemon" ) ) {

	    $suffix = '_' . ++$seq;
	    die "Unable to find a free name for $inidir/${Wikiname}TaskDaemon after $seq attempts\n"
	      if $seq >= 100;
	}
	$ininame = "${Wikiname}TaskDaemon$suffix";
    }

    unless( 
	   # Startup (system boot) link for init
	   slink( "$tasklib/Daemon", "$inidir/$ininame" ) &&

	   # Following links will be followed from "$inidir/$ininame", but those won't
	   # exist if we're running -n.  Since they get realpathed, using the eventual
	   # target has the same effect.

	   # Link for startup to find bin/
	   slink( $bindir, realpath( "$tasklib/Daemon" ) . '_bin' ) &&

	   # Link from Daemon to selected ini file name for uninstall

	   slink( "$inidir/$ininame", realpath( "$tasklib/Daemon" ) . '_init' )
	  ) {
	print STDERR "Unable to register daemon startup links\n";
	return 0;
    }

    return 1;
}

sub slink {
    my( $old, $new ) = @_;

    # $new is the symlink being created, $old is what will be the target of the symlink
    # Try to create a relative path from the new back to the target.
    #
    # This is ugly because abs2rel/rel2abs want just directory specs.

    my $relTgt = $old;
    $relTgt .= '/' if -d $old;

    $relTgt = File::Spec->abs2rel( (fileparse($relTgt))[1], (fileparse($new))[1] ) . (-d $old? '' : '/' . (fileparse($relTgt))[0] );

    print "    symlink: $new => $relTgt  " if $verbose;

    # If the desired link already exists
    #  If it points to where we want, we're done.
    #  Otherwise, we can remove it -- only if -f was used.

    if( -l $new ) {
	# Existing link is relative to its location; make absolute so we can realpath it
	my $oldLnk = readlink( $new  );

	if( $relTgt eq $oldLnk ) {
	    print "[ok] - no change\n" if $verbose;
	    return 1;
	}

	print "    $new => $old  " unless $verbose;

	print "exists => " . readlink($new);

	unless( $cliOptions{f} ) {
	    print " -f required to remove\n";
	    return 0;
	}
	if( $cliOptions{n} ) {
	    print " [Would remove $new]\n";
	    return 1;
	}
	delete $cliOptions{f};

	print " [Removing $new";
	unless( unlink( $new ) ) {
	    print " - failed: $!\n";
	    return 0;
	}
	print " - OK]";
    }

    if( $cliOptions{n} ) {
	print "\n" if $verbose;
	return 1;
    }

    if( symlink $relTgt, $new ) {
	print "[OK]\n" if $verbose;
	return 1;
    }
    print "    $new => $relTgt  " unless $verbose;
    print "Failed: $!\n";
    return 0;
}

sub sunlink {
    my( $link ) = @_;

    unless( -l $link ) {
	print "$link: Not a symlink\n";
	return 0;
    }

    print "    Removing symlink $link => " . readlink($link) if $verbose;

    if( $cliOptions{n} ) {
	print "\n" if $verbose;
	return 1;
    }

    if( unlink $link ) {
	print "[OK]\n" if $verbose;
	return 1;
    }
    print "    Remove $link  " unless $verbose;
    print "Failed: $!\n";
    return 0;
}

sub checkwiki {

    unless( -d $bindir && -x "$bindir/configure" && 
	    -x "$bindir/view" && -x "$bindir/edit" && 
	    -r "$bindir/setlib.cfg" ) {
	print STDERR "$bindir does not appear to be a wiki bin directory\n";
	return 0;
    }
    my $wd = getcwd();
    chdir File::Spec->catfile( $bindir, File::Spec->updir ) or die "Chdir:$!\n";
    $rootdir = realpath( '.' );
    unshift @INC, '.', $bindir;
    eval "
	our \$${wikiname}LibPath;
	unless( my \$sts = do \"$bindir/setlib.cfg\" ) {
	    die \"Couldn't parse $bindir/setlib.cfg: \$@\\n\" if( \$@ );
	    die \"Couldn't read $bindir/setlib.cfg: \$!\\n\" unless defined \$sts;
	    die \"Unsuccessful status from $bindir/setlib.cfg: \$sts\\n\" unless \$sts;
            die \"lost\";
	}
	die \"$bindir/setlib.cfg does not contain a valid \\\$${wikiname}LibPath\\n\"
            unless( \$${wikiname}LibPath && -r \"\$${wikiname}LibPath/LocalSite.cfg\" );
        \$libdir = \$${wikiname}LibPath;
    "; if( $@ ) {
	chdir $wd or die "chdir:$!\n";
	print STDERR "$bindir does not contain a valid setlib.cfg:$@\n";
	return 0;
    }
    $libdir .= '/' unless $libdir =~ m./$.;
    $libdir = File::Spec->rel2abs( (fileparse($libdir))[1], $bindir );
#    $libdir .= '/' unless $libdir =~ m./$.;
    $libdir = realpath( $libdir );

    $wikiver = `grep '\$RELEASE = ' $libdir/$Wikiname.pm`;
    $wikiver =~ s/^.*?\$RELEASE = '([^']*)'.*$/$1/ms;
    print "$Wikiname ($wikiver) platform detected\n" if $verbose;

    eval "
	unless( my \$sts = do \"$libdir/LocalSite.cfg\" ) {
	    die \"Couldn't parse $libdir/LocalSite.cfg: \$@\\n\" if( \$@ );
	    die \"Couldn't read $libdir/LocalSite.cfg: \$!\\n\" unless defined \$sts;
	    die \"Unsuccessful status from $libdir/LocalSite.cfg: \$sts\\n\" unless \$sts;
            die \"lost\";
        }
# Not in TWiki 4/5
#	die \"$libdir/LocalSite.cfg does not contain a valid \\\%${Wikiname}::cfg\\n\"
#            unless \$${Wikiname}::cfg{ScriptDir} && realpath( \$${Wikiname}::cfg{ScriptDir} ) eq \$bindir;
        \$cfg = \\\%${Wikiname}::cfg;
    "; if( $@ || !( $cfg->{PubDir} && -d $cfg->{PubDir} &&
	    $cfg->{DataDir} && -d $cfg->{DataDir} &&
	    $cfg->{TemplateDir} && -d $cfg->{TemplateDir} &&
	    $cfg->{LocalesDir} && -d $cfg->{LocalesDir} &&
# Not in TWiki 4/5
#	    $cfg->{ToolsDir} && -d $cfg->{ToolsDir} &&
	    $cfg->{WorkingDir} && -d $cfg->{WorkingDir} ) ) {
	   chdir $wd or die "chdir:$!\n";
	   print STDERR "$libdir does not contain a valid LocalSite.cfg:$@\n";
	   return 0;
    }
    chdir $wd or die "chdir:$!\n";

    $wikilib = "$libdir/$Wikiname";
    $tasklib  = "$wikilib/Tasks";

    unless( -r "$libdir/$Wikiname.pm" && 
	    -r "$wikilib/Plugins/TasksPlugin.pm" && 
	    -x "$wikilib/Tasks/Daemon" && -r "$wikilib/Plugins/TasksPlugin/MANIFEST" ) {
	print STDERR "$libdir does not appear to be a $Wikiname lib directory (or the installation is incomplete\n";
	return 0;
    }

    if( exists $cfg->{ToolsDir} ) { # Believe config if present ( currently only Foswiki)
	print STDERR "-t ignored because tools directory found in $Wikiname configuration\n" if( $cliOptions{t} );
	$toolsdir = $cfg->{ToolsDir};
    }
    unless( -d $toolsdir && -x "$toolsdir/tick_$wikiname.pl" && (-f "$toolsdir/tick_$wikiname.pl" || -l "$toolsdir/tick_$wikiname.pl")  && -r "$toolsdir/extender.pl" ) {
	print STDERR "$toolsdir does not appear to be a $Wikiname tools directory\n";
	return 0;
    }

    if( $verbose ) {
	print "Using the following directories for installation:\n",
	"    bin: $bindir => " . realpath( $bindir ) . "\n",
	"    lib: $libdir => " . realpath( $libdir ) . "\n",
	"  tools: $toolsdir => " . realpath( $toolsdir ) . "\n";
    }

    return 1;
}

sub checkinit {
    unless( defined $inidir ) {
	print STDERR "Autostart will not be enabled\n";
	return 1;
    }

    unless( -d $inidir && -w $inidir ) {
	print STDERR "$inidir does not appear to be valid and writable\n";
	return 0;
    }
    return 1;
}

sub getmappings {
    my( $mapfile, $kitmap, $libmap ) = @_;

      # File specified with -m

      if( $mapfile ) {
	# Explictly specified file
	readmapfile( $mapfile ) or die "Failed to read $mapfile: $!\n";
	return $mapfile;
    }

    # See if there is a site policy file in $libdir

    $mapfile = "$libmap.LocalSite";
    if( -f $mapfile ) {
	readmapfile( $mapfile ) or die "Failed to read $mapfile: $!\n";
	return $mapfile;
    }

    # Look for a distribution-specific file

    # First, get OS (or distribution) name & version

    my $distname = $^O;
    my $distver = '';

    if( $distname eq 'linux' ) {
	# linux requires extra work...

	require Linux::Distribution;
	$distname =  Linux::Distribution::distribution_name();
	unless( $distname ) {
	    print STDERR "Unable to determine distribution name; update Linux::Distribution or use -m to specify manifest mapping file\n";
	    exit 1;
	}
	$distver = Linux::Distribution::distribution_version();
	unless( $distver ) {
	    if( $distname eq 'fedora' ) {
		print STDERR "Can't determine fedora distribution's version; see https://rt.cpan.org/Public/Bug/Display.html?id=69671 for a patch\n";
	    } else {
		print STDERR "Unable to determine $distname version - is lsb package (containing lsb_release command and data) installed?\nWill only look for generic $distname mapping\n";
	    }
	}
    }

    # Try the current kit's MANIFEST's directory, then the wiki library
    foreach my $mapbase ($kitmap, $libmap) {
	# First, distribution.version
	$mapfile = "$mapbase.$distname.$distver";
	if( $distver && -f $mapfile ) {
	    readmapfile( $mapfile ) or die "Failed to read $mapfile: $!\n";
	    return $mapfile;
	}
	# Use generic distribution
	$mapfile = "$mapbase.$distname";
	if( -f $mapfile ) {
	    readmapfile( $mapfile ) or die "Failed to read $mapfile: $!\n";
	    return $mapfile;
	}
    }
    # No map file found
    if( $selinux ) {
	print STDERR "Selinux install requires a manifest mapping file, which is not present.  Obtain mapping file, or use -z to install without setting selinux types\n";
	exit 1;
    }
    print "No mapping file found, proceeding with defaults\n" if( $verbose );
    return '';
}

sub readmapfile {
    my $mapfile = shift;

    eval {
	unless( my $sts = do $mapfile ) {
	    die "Couldn't parse $mapfile: $@\n" if( $@ );
	    die "Couldn't read $mapfile: $!" unless( defined $sts );
	    die "Unsuccessful status from $mapfile\n" unless( $sts );
	}
    }; if( $@ ) {
	print STDERR "Unable to process configuration map: $@\n";
	exit 1;
    }

    print STDERR "Obtained configuration from $mapfile\n" if( $verbose );
    return 1;
}

sub getperms {
    my( $manfile ) = @_;

    # Extract permissions from the manifest file
    # Structured comments provide additional information
    # ##S foo_t - selinux context for following files (just type or whole specifier)
    # ##U user.group - to be applied to file
    # ##D dirname mode desc - permissions for directories (not traditionally in MANIFEST)
    #
    # All permissions and users can be mapped through a site or distribution-specific
    # configuration file.  This is required to translate the generic wiki selinux
    # types used in MANIFEST to provide distribution independence.  It's optional
    # for mode, user and group to allow for site policy.
    # The mappings have already been read from $mapfile

    print STDERR "Processing $manfile\n" if $verbose;

    # Would be nice to validate security contexts, but we use older type names for maximum compatibility
    # and these are aliases in current releases.  And seinfo won't list aliases (nor will anything else
    # identified so far... If we had a list we could validate ##S lines and save a lot of chcon errors.

    my %sectxlist;
    if( 0&& $selinux && open( my $ctx, '-|', 'seinfo -t' ) ) {
	while( <$ctx> ) {
	    chomp;
	    s/^\s+//;
	    s/\s+$//;
	    $sectxlist{$_} = 1;
	}
	close $ctx;
    }

    open( my $fh, '<', $manfile ) or die "Can't open $manfile: $!\n";
    my $perms = {};
    my( $sctxu, $sctxt, $user, $group ) = ( '', '', '', '');

    while( <$fh> ) {
	if( /^##S\s+([\w_-]+)\s*(?:#.*)?$/ ) { # E.g. wiki_topic_t
	    # Security context type introducer
	    # Map from generic to distribution and save for following files

	    next unless( $selinux );

	    $sctxt = $selinuxTypeMap{$1};
	    unless( defined $sctxt ) {
		print STDERR "No mapping in $mapfile for $1, please update file\n";
		exit 1;
	    }
	    $sctxu = $selinuxUser;
	    next;
	}
	if( /^##U\s+([\w_-]+)\.([\w_-]+)\s*(?:#.*)?$/ ) { # E.g. webserverUser.webserverGroup
	    # File ownership introducer
	    # Map from generic name to distribution (or default) & save for following files

	    ( $user, $group ) = ($userGroupMap{$1}, $userGroupMap{$2} );
	    unless( defined $user ) {
		print STDERR "No mapping for user $1 in $mapfile\n";
		exit 1;
	    }
	    unless( defined $group ) {
		print STDERR "No mapping for group $2 in $mapfile\n";
		exit 1;
	    }

	    # Convert names to uid/gid & cache

	    $users{$user} = getpwnam( $user ) unless exists $users{$user};
	    print STDERR "Unknown user '$user' specifed in $manfile line $., ignoring\n" unless defined $users{$user};
	    $groups{$group} = getgrnam( $group ) unless exists $groups{$group};
	    print STDERR "Unknown group '$group' specifed in $manfile line $., ignoring\n" unless defined $groups{$group};
	    next;
	}

	my( $file, $perm, $desc );
	# Directory specifier - not in traditional MANIFEST
	# Treat as file for setting permissions
	unless( ($file, $perm, $desc) = /^##D\s+(\S+)\s+(\d+)(?:\s+(.*))?$/ ) {
	    # Strip comments, blank lines and leading/trailing spaces
	    s/#.*$//;
	    s/^\s+//;
	    s/\s+$//;
	    next unless length;

	    if( /^!include\s+(.*)$/ ) {
# !include seems to have ../ (sometimes) + package name (huh?)/
# Not sure what to do...For now, at least we can skip these.
#		$perms = { %$perms,
#			   %{ getperms( (fileparse( $manfile ))[1] . $1 ) },
#			 };
#		my $incfile = $1;
#		$incfile =~ s!^(?:\.\./)?\w+/!!;
#		$perms = { %$perms,
#			   %{ getperms( "$rootdir/$incfile" ) },
#			 };
#		print STDERR "Processing continues in $manfile\n" if $verbose;
		next;
	    }

	    ($file, $perm, $desc) = split( /\s+/, $_, 3 );
	}

	# Record filename, mode, security context, and owner

	next unless( $file );

	if( defined $perm ) { 
	    $perm = oct( $perm ) if $perm =~ /^0/;
	} else {
	    # apply defaults - documented in BuildContrib
	    if( -d "$rootdir/$file" ) {
		$perm = 0775;
	    } elsif( $file =~ /\.pm$/ ) {
		$perm = 0444;
	    } elsif( $file =~ /\.pl$/ ) {
		$perm = 0554;
	    } elsif( $file =~ m!^data/.*\.txt$! ) {
		$perm = 0664;
	    } elsif( $file =~ m!^pub/! ) {
		$perm = 0664;
	    } elsif( $file =~ m!^bin/! ) {
		$perm = 0555;
	    } else {
		$perm = 0444;
	    }
	}

	@{$perms->{$file}}{'mode','sctxu','sctxt','user','group'} = ($perm,$sctxu,$sctxt,$users{$user},$groups{$group});
    }
    close $fh or die "Error closing: $manfile:$!\n";

    return $perms;
}

sub setperms {
    my $root = shift;
    my $filename = shift;

    return 0 unless exists $perms->{$filename};

    my $file = File::Spec->catfile( $root,$filename );
    return 0 unless( -e $file );

    my( $user, $group, $mode, $sctxu, $sctxt ) = @{$perms->{$filename}}{'user','group','mode','sctxu', 'sctxt'};

    if( defined $user || defined $group ) {
	$user = -1 unless( defined $user );
	$group = -1 unless( defined $group );

	if( $verbose ) {
	    my( $curu, $curg ) =  (stat($file))[4,5];
	    # (-l $file? (lstat($file))[4,5] : (stat($file))[4,5]);
	    printf "%s %s.%s -> %s.%s\n", $file, $curu, $curg, ($user == -1? '*' : $user), ($group == -1? '*' : $group) if( ($user != -1 && $user != $curu) || ($group != -1 && $group != $curg) );
	}
	unless( $cliOptions{n} ) {
	    if( -l $file ) {
		lchown $user,$group, $file or warn "lchown $file: $!\n";
	    }
	    chown $user,$group, $file or warn "chown $file: $!\n";
	}
    }

    if( defined $mode && !$permissiveOK ) {
	# Handle any permission mapping and masking

	if( -d $file ) {
	    $mode = $dirPermMap{$mode} if( exists $dirPermMap{$mode} );
	    $mode &= ~$dirUmask;
	} else {
	    $mode = $filePermMap{$mode} if( exists $filePermMap{$mode} );
	    $mode &= ~$fileUmask;
	}

	if( $verbose ) {
	    my $curperm = (stat( $file ))[2] & 07777;
	    printf "$file %03lo -> %03lo\n", $curperm, $mode if( $mode != $curperm );
	}

	unless( $cliOptions{n} ) {
	    chmod $mode, $file or warn "chmod $file:$!\n";
	}
    }

    if( defined $sctxt && $selinux ) {
	# Type and range from mapped type
	my @sctx = split( ':', $sctxt,2 );
	my $sctx = "-t $sctx[0]" if( $sctx[0] );
	$sctx = "$sctx -l $sctx[1]" if( $sctx[1] );
	# User if specified
	$sctx = "-u $sctxu $sctx" if( $sctxu );
	die "Bad security context mapping: $sctxt\n" unless( defined $sctx );
	if( $verbose ) {
	    # Note that this doesn't account for aliases, so will be noisy if new and old are aliased.
	    # Also note that stat is an external command, so verbosity isn't cheap.
	    my @cursec = split( ':', `stat -c '%C' $file` );
	    my @newsec = @cursec;
	    $newsec[0] = $sctxu if( $sctxu );
	    # Role never touched (a file is an object everywhere...)
	    $newsec[2] = $sctx[0] if( $sctx[0] );
	    $newsec[3] = $sctx[1] if( $sctx[1] );
	    my $csec = join(':', @cursec); chomp $csec;
	    my $nsec = join(':', @newsec); chomp $nsec;
	    printf "$file %s -> %s (%s)\n", $csec, $nsec, $sctx if($nsec ne $csec);
	}
	# At this time, there is no known perl library/module for chcon; must use system...
	unless( $cliOptions{n} ) {
	    if( -l $file ) {
		system( "chcon -h $sctx $file" ) == 0 or warn "chcon -h $file failed\n";
	    }
	    system( "chcon $sctx $file" ) == 0 or warn "chcon $file failed\n";
	}
    }
}

sub HELP_MESSAGE {
    my( $ofh, $optpkg, $optver, $sws ) = @_;

    my $pname = $FindBin::Script;

    print $ofh <<"USAGE" ;
$pname (un)installation for $Wikiname tasking framework

Usage:
  $pname options command

Options:
  --help    - print usage summary
  -v        - Detail messages for all actions
  -f        - Remove existing symlinks that conflict
  -n        - Don\'t make any changes
  -b dir    - specify location of $Wikiname bin directory
              typically <webdocroot>/$wikiname/bin
  -t dir    - specify location of $Wikiname tools directory
              typically <webdocroot>/$wikiname/tools (Foswiki
              can get this automagically from config file)
  -s dir    - specify location of system startup directory
              typically /etc/init.d, or '-' if autostart 
              is not desired.
  -m mapfile  Use mapfile to translate MANIFEST permissions to installed.
              MANIFEST permissions are generic and generally overly permissive.
              The mapfile enables you to specify a different site policy, and
              allows the generic permissions to be adapted to specific distributions' requirements.
              Default is lib/MANIFEST.map.LocalSite, or
              lib/TWiki/Plugins/TasksPlugin/MANIFEST.map.<distribution-name>.<version> or
              lib/TWiki/Plugins/TasksPlugin/MANIFEST.map.<distribution-name> or
              lib/MANIFEST.map.<distribution-name>.<version>, or
              lib/MANIFEST.map.<distribution-name>, or
              hardwired defaults (SeLinux requires file)
  -p          Leave permissive permissions as specified in MANIFEST
  -u 0nnn     Restrict file permissions using this umask (default: 007)
  -U 0nnn     Restrict directory permissions umask (default: 006)
  -Z          Set SeLinux security context on files specified in MANIFEST
              Default if SeLinux detected.
  -z          Do not set SeLinux security contexts even if SeLinux is detected.

Commands:
   install  - Install symbolic links
   uninstall- Remove symbolic links
USAGE
    return;
}
__END__
# Utility for TWiki/Foswiki platforms
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at
# http://www.gnu.org/copyleft/gpl.html
